/**
 * {颜色处理}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/03/06 13:05:09
 */

import { cacher, is } from '../utils.js'
import { COMMA_SPACES } from './constants.js'

const hsrg = { hs: 1, rg: 1 }
const colourRegExp =
  /^\s*((#[a-f\d]{6})|(#[a-f\d]{3})|rgba?\(\s*([\d\.]+%?\s*,\s*[\d\.]+%?\s*,\s*[\d\.]+%?(?:\s*,\s*[\d\.]+%?)?)\s*\)|hsba?\(\s*([\d\.]+(?:deg|\xb0|%)?\s*,\s*[\d\.]+%?\s*,\s*[\d\.]+(?:%?\s*,\s*[\d\.]+)?%?)\s*\)|hsla?\(\s*([\d\.]+(?:deg|\xb0|%)?\s*,\s*[\d\.]+%?\s*,\s*[\d\.]+(?:%?\s*,\s*[\d\.]+)?%?)\s*\))\s*$/i

function toHex(color) {
  let i =
      doc.getElementsByTagName('head')[0] || doc.getElementsByTagName('svg')[0],
    red = 'rgb(255, 0, 0)'
  toHex = cacher(function (color) {
    if (color.toLowerCase() == 'red') {
      return red
    }
    i.style.color = red
    i.style.color = color
    let out = doc.defaultView.getComputedStyle(i, E).getPropertyValue('color')
    return out == red ? null : out
  })
  return toHex(color)
}
function hsbtoString() {
  return 'hsb(' + [this.h, this.s, this.b] + ')'
}
function hsltoString() {
  return 'hsl(' + [this.h, this.s, this.l] + ')'
}
function rgbtoString() {
  return this.opacity == 1 || this.opacity == null
    ? this.hex
    : 'rgba(' + [this.r, this.g, this.b, this.opacity] + ')'
}
function prepareRGB(r, g, b) {
  if (g == null && is(r, 'object') && 'r' in r && 'g' in r && 'b' in r) {
    b = r.b
    g = r.g
    r = r.r
  }
  if (g == null && is(r, string)) {
    let clr = getRGB(r)
    r = clr.r
    g = clr.g
    b = clr.b
  }
  if (r > 1 || g > 1 || b > 1) {
    r /= 255
    g /= 255
    b /= 255
  }

  return [r, g, b]
}
function packageRGB(r, g, b, o) {
  r = Math.round(r * 255)
  g = Math.round(g * 255)
  b = Math.round(b * 255)
  let rgb = {
    r: r,
    g: g,
    b: b,
    opacity: is(o, 'finite') ? o : 1,
    hex: rgb(r, g, b),
    toString: rgbtoString
  }
  is(o, 'finite') && (rgb.opacity = o)
  return rgb
}

// Colour

export const getRGB = cacher(function (colour) {
  if (!colour || !!((colour = String(colour)).indexOf('-') + 1)) {
    return {
      r: -1,
      g: -1,
      b: -1,
      hex: 'none',
      error: 1,
      toString: rgbtoString
    }
  }
  if (colour == 'none') {
    return { r: -1, g: -1, b: -1, hex: 'none', toString: rgbtoString }
  }
  !(
    hsrg.hasOwnProperty(colour.toLowerCase().slice(0, 2)) ||
    colour.charAt() == '#'
  ) && (colour = toHex(colour))

  if (!colour) {
    return {
      r: -1,
      g: -1,
      b: -1,
      hex: 'none',
      error: 1,
      toString: rgbtoString
    }
  }
  let res,
    red,
    green,
    blue,
    opacity,
    t,
    values,
    rgb = colour.match(colourRegExp)
  if (rgb) {
    if (rgb[2]) {
      blue = parseInt(rgb[2].substring(5), 16)
      green = parseInt(rgb[2].substring(3, 5), 16)
      red = parseInt(rgb[2].substring(1, 3), 16)
    }
    if (rgb[3]) {
      blue = parseInt((t = rgb[3].charAt(3)) + t, 16)
      green = parseInt((t = rgb[3].charAt(2)) + t, 16)
      red = parseInt((t = rgb[3].charAt(1)) + t, 16)
    }
    if (rgb[4]) {
      values = rgb[4].split(COMMA_SPACES)
      red = +values[0]
      values[0].slice(-1) == '%' && (red *= 2.55)
      green = +values[1]
      values[1].slice(-1) == '%' && (green *= 2.55)
      blue = +values[2]
      values[2].slice(-1) == '%' && (blue *= 2.55)
      rgb[1].toLowerCase().slice(0, 4) == 'rgba' && (opacity = +values[3])
      values[3] && values[3].slice(-1) == '%' && (opacity /= 100)
    }
    if (rgb[5]) {
      values = rgb[5].split(COMMA_SPACES)
      red = +values[0]
      values[0].slice(-1) == '%' && (red /= 100)
      green = +values[1]
      values[1].slice(-1) == '%' && (green /= 100)
      blue = +values[2]
      values[2].slice(-1) == '%' && (blue /= 100)
      ;(values[0].slice(-3) == 'deg' || values[0].slice(-1) == '\xb0') &&
        (red /= 360)
      rgb[1].toLowerCase().slice(0, 4) == 'hsba' && (opacity = +values[3])
      values[3] && values[3].slice(-1) == '%' && (opacity /= 100)
      return hsb2rgb(red, green, blue, opacity)
    }
    if (rgb[6]) {
      values = rgb[6].split(COMMA_SPACES)
      red = +values[0]
      values[0].slice(-1) == '%' && (red /= 100)
      green = +values[1]
      values[1].slice(-1) == '%' && (green /= 100)
      blue = +values[2]
      values[2].slice(-1) == '%' && (blue /= 100)
      ;(values[0].slice(-3) == 'deg' || values[0].slice(-1) == '\xb0') &&
        (red /= 360)
      rgb[1].toLowerCase().slice(0, 4) == 'hsla' && (opacity = +values[3])
      values[3] && values[3].slice(-1) == '%' && (opacity /= 100)
      return hsl2rgb(red, green, blue, opacity)
    }
    red = Math.min(Math.round(red), 255)
    green = Math.min(Math.round(green), 255)
    blue = Math.min(Math.round(blue), 255)
    opacity = Math.min(Math.max(opacity, 0), 1)
    rgb = { r: red, g: green, b: blue, toString: rgbtoString }
    rgb.hex =
      '#' + (16777216 | blue | (green << 8) | (red << 16)).toString(16).slice(1)
    rgb.opacity = is(opacity, 'finite') ? opacity : 1
    return rgb
  }
  return { r: -1, g: -1, b: -1, hex: 'none', error: 1, toString: rgbtoString }
})
/*\
 **
 * Converts HSB values to a hex representation of the color
 - h (number) hue
 - s (number) saturation
 - b (number) value or brightness
 = (string) hex representation of the color
\*/
export const hsb = cacher(function (h, s, b) {
  return hsb2rgb(h, s, b).hex
})
/*\
 **
 * Converts HSL values to a hex representation of the color
 - h (number) hue
 - s (number) saturation
 - l (number) luminosity
 = (string) hex representation of the color
\*/
export const hsl = cacher(function (h, s, l) {
  return hsl2rgb(h, s, l).hex
})
/*\
 **
 * Converts RGB values to a hex representation of the color
 - r (number) red
 - g (number) green
 - b (number) blue
 = (string) hex representation of the color
\*/
export const rgb = cacher(function (r, g, b, o) {
  if (is(o, 'finite')) {
    let round = Math.round
    return 'rgba(' + [round(r), round(g), round(b), +o.toFixed(2)] + ')'
  }
  return '#' + (16777216 | b | (g << 8) | (r << 16)).toString(16).slice(1)
})

/*\
 **
 * Parses the color string and returns an object featuring the color's component values
 - clr (string) color string in one of the supported formats (see getRGB)
 = (object) Combined RGB/HSB object in the following format:
 o {
 o     r (number) red,
 o     g (number) green,
 o     b (number) blue,
 o     hex (string) color in HTML/CSS format: #••••••,
 o     error (boolean) `true` if string can't be parsed,
 o     h (number) hue,
 o     s (number) saturation,
 o     v (number) value (brightness),
 o     l (number) lightness
 o }
\*/
export const parseColor = function (clr) {
  let rgb
  if (is(clr, 'object') && 'h' in clr && 's' in clr && 'b' in clr) {
    rgb = hsb2rgb(clr)
    clr.r = rgb.r
    clr.g = rgb.g
    clr.b = rgb.b
    clr.opacity = 1
    clr.hex = rgb.hex
  } else if (is(clr, 'object') && 'h' in clr && 's' in clr && 'l' in clr) {
    rgb = hsl2rgb(clr)
    clr.r = rgb.r
    clr.g = rgb.g
    clr.b = rgb.b
    clr.opacity = 1
    clr.hex = rgb.hex
  } else {
    if (is(clr, 'string')) {
      clr = getRGB(clr)
    }
    if (
      is(clr, 'object') &&
      'r' in clr &&
      'g' in clr &&
      'b' in clr &&
      !('error' in clr)
    ) {
      rgb = rgb2hsl(clr)
      clr.h = rgb.h
      clr.s = rgb.s
      clr.l = rgb.l
      rgb = rgb2hsb(clr)
      clr.v = rgb.b
    } else {
      clr = { hex: 'none' }
      clr.r = clr.g = clr.b = clr.h = clr.s = clr.v = clr.l = -1
      clr.error = 1
    }
  }
  clr.toString = rgbtoString
  return clr
}
/*\
 **
 * Converts HSB values to an RGB object
 - h (number) hue
 - s (number) saturation
 - v (number) value or brightness
 = (object) RGB object in the following format:
 o {
 o     r (number) red,
 o     g (number) green,
 o     b (number) blue,
 o     hex (string) color in HTML/CSS format: #••••••
 o }
\*/
export const hsb2rgb = function (h, s, v, o) {
  if (is(h, 'object') && 'h' in h && 's' in h && 'b' in h) {
    v = h.b
    s = h.s
    o = h.o
    h = h.h
  }
  h *= 360
  let R, G, B, X, C
  h = (h % 360) / 60
  C = v * s
  X = C * (1 - abs((h % 2) - 1))
  R = G = B = v - C

  h = ~~h
  R += [C, X, 0, 0, X, C][h]
  G += [X, C, C, X, 0, 0][h]
  B += [0, 0, X, C, C, X][h]
  return packageRGB(R, G, B, o)
}
/*\
 **
 * Converts HSL values to an RGB object
 - h (number) hue
 - s (number) saturation
 - l (number) luminosity
 = (object) RGB object in the following format:
 o {
 o     r (number) red,
 o     g (number) green,
 o     b (number) blue,
 o     hex (string) color in HTML/CSS format: #••••••
 o }
\*/
export const hsl2rgb = function (h, s, l, o) {
  if (is(h, 'object') && 'h' in h && 's' in h && 'l' in h) {
    l = h.l
    s = h.s
    h = h.h
  }
  if (h > 1 || s > 1 || l > 1) {
    h /= 360
    s /= 100
    l /= 100
  }
  h *= 360
  let R, G, B, X, C
  h = (h % 360) / 60
  C = 2 * s * (l < 0.5 ? l : 1 - l)
  X = C * (1 - abs((h % 2) - 1))
  R = G = B = l - C / 2

  h = ~~h
  R += [C, X, 0, 0, X, C][h]
  G += [X, C, C, X, 0, 0][h]
  B += [0, 0, X, C, C, X][h]
  return packageRGB(R, G, B, o)
}
/*\
 **
 * Converts RGB values to an HSB object
 - r (number) red
 - g (number) green
 - b (number) blue
 = (object) HSB object in the following format:
 o {
 o     h (number) hue,
 o     s (number) saturation,
 o     b (number) brightness
 o }
\*/
export const rgb2hsb = function (r, g, b) {
  b = prepareRGB(r, g, b)
  r = b[0]
  g = b[1]
  b = b[2]

  let H, S, V, C
  V = Math.max(r, g, b)
  C = V - Math.min(r, g, b)
  H =
    C == 0
      ? null
      : V == r
      ? (g - b) / C
      : V == g
      ? (b - r) / C + 2
      : (r - g) / C + 4
  H = (((H + 360) % 6) * 60) / 360
  S = C == 0 ? 0 : C / V
  return { h: H, s: S, b: V, toString: hsbtoString }
}
/*\
 **
 * Converts RGB values to an HSL object
 - r (number) red
 - g (number) green
 - b (number) blue
 = (object) HSL object in the following format:
 o {
 o     h (number) hue,
 o     s (number) saturation,
 o     l (number) luminosity
 o }
\*/
export const rgb2hsl = function (r, g, b) {
  b = prepareRGB(r, g, b)
  r = b[0]
  g = b[1]
  b = b[2]

  let H, S, L, M, m, C
  M = Math.max(r, g, b)
  m = Math.min(r, g, b)
  C = M - m
  H =
    C == 0
      ? null
      : M == r
      ? (g - b) / C
      : M == g
      ? (b - r) / C + 2
      : (r - g) / C + 4
  H = (((H + 360) % 6) * 60) / 360
  L = (M + m) / 2
  S = C == 0 ? 0 : L < 0.5 ? C / (2 * L) : C / (2 - 2 * L)
  return { h: H, s: S, l: L, toString: hsltoString }
}
